from __future__ import annotations

import re
from collections.abc import Callable

import numpy as np
import pandas as pd
import pytest
from pytest_mock import MockerFixture

from tests.framework.results.helpers import (
    HOUSE_CATEGORIES,
    NAME,
    NAME_COLUMNS,
    STUDENT_HOUSES,
    STUDENT_TABLE,
    sorting_hat_bad_mapping,
    sorting_hat_serial,
    sorting_hat_vectorized,
)
from vivarium.framework.results.manager import ResultsManager
from vivarium.framework.results.stratification import (
    Stratification,
    get_mapped_col_name,
    get_original_col_name,
)


@pytest.mark.parametrize(
    "mapper, is_vectorized",
    [
        (  # expected output for vectorized
            sorting_hat_vectorized,
            True,
        ),
        (  # expected output for non-vectorized
            sorting_hat_serial,
            False,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_stratification(
    mapper: Callable[[pd.DataFrame], pd.Series[str]] | Callable[[pd.Series[str]], str],
    is_vectorized: bool,
) -> None:
    my_stratification = Stratification(
        name=NAME,
        sources=NAME_COLUMNS,
        categories=HOUSE_CATEGORIES,
        excluded_categories=[],
        mapper=mapper,
        is_vectorized=is_vectorized,
    )
    output = my_stratification.stratify(STUDENT_TABLE)
    assert isinstance(output, pd.Series)
    assert output.eq(STUDENT_HOUSES).all()


@pytest.mark.parametrize(
    "sources, categories, mapper, msg_match",
    [
        (
            [],
            HOUSE_CATEGORIES,
            None,
            (
                f"No mapper but 0 stratification sources are provided for stratification {NAME}. "
                "The list of sources must be of length 1 if no mapper is provided."
            ),
        ),
        (
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            None,
            (
                f"No mapper but {len(NAME_COLUMNS)} stratification sources are provided for stratification {NAME}. "
                "The list of sources must be of length 1 if no mapper is provided."
            ),
        ),
        (
            [],
            HOUSE_CATEGORIES,
            sorting_hat_vectorized,
            "The sources argument must be non-empty.",
        ),
        (
            NAME_COLUMNS,
            [],
            FileNotFoundError,
            "The categories argument must be non-empty.",
        ),
    ],
    ids=[
        "no_mapper_empty_sources",
        "no_mapper_multiple_sources",
        "with_mapper_empty_sources",
        "empty_categories",
    ],
)
def test_stratification_init_raises(
    sources: list[str],
    categories: list[str],
    mapper: Callable[[pd.DataFrame], pd.Series[str]] | Callable[[pd.Series[str]], str],
    msg_match: str,
) -> None:
    with pytest.raises(ValueError, match=re.escape(msg_match)):
        Stratification(NAME, sources, categories, [], mapper, True)


@pytest.mark.parametrize(
    "sources, mapper, is_vectorized, expected_exception, error_match",
    [
        (
            NAME_COLUMNS,
            sorting_hat_bad_mapping,
            False,
            ValueError,
            "Invalid values mapped to hogwarts_house: {'pancakes'}",
        ),
        (
            ["middle_initial"],
            sorting_hat_vectorized,
            True,
            KeyError,
            "None of [Index(['middle_initial'], dtype='object')] are in the [columns]",
        ),
        (
            NAME_COLUMNS,
            sorting_hat_serial,
            True,
            Exception,  # Can be any exception
            "",  # Can be any error message
        ),
        (
            NAME_COLUMNS,
            sorting_hat_vectorized,
            False,
            Exception,  # Can be any exception
            "",  # Can be any error message
        ),
        (
            NAME_COLUMNS,
            lambda df: pd.Series(np.nan, index=df.index),
            True,
            ValueError,
            f"Invalid values mapped to hogwarts_house: {{{np.nan}}}",
        ),
    ],
    ids=[
        "unknown_category",
        "source_not_in_population_columns",
        "vectorized_with_serial_mapper",
        "not_vectorized_with_vectorized_mapper",
        "mapper_returns_null",
    ],
)
def test_stratification_call_raises(
    sources: list[str],
    mapper: Callable[[pd.DataFrame], pd.Series[str]] | Callable[[pd.Series[str]], str],
    is_vectorized: bool,
    expected_exception: type[Exception],
    error_match: str,
) -> None:
    my_stratification = Stratification(
        NAME, sources, HOUSE_CATEGORIES, [], mapper, is_vectorized
    )
    with pytest.raises(expected_exception, match=re.escape(error_match)):
        my_stratification.stratify(STUDENT_TABLE)


@pytest.mark.parametrize("default_stratifications", [["age", "sex"], ["age"], []])
def test_setting_default_stratifications(
    default_stratifications: list[str], mocker: MockerFixture
) -> None:
    """Test that default stratifications are set as expected."""
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification.default = default_stratifications

    mgr.setup(builder)

    assert mgr._results_context.default_stratifications == default_stratifications


def test_get_mapped_column_name() -> None:
    assert get_mapped_col_name("foo") == "foo_mapped_values"


@pytest.mark.parametrize(
    "col_name, expected",
    [
        ("foo_mapped_values", "foo"),
        ("foo", "foo"),
        ("foo_mapped_values_mapped_values", "foo_mapped_values"),
        ("foo_mapped_values2", "foo_mapped_values2"),
        ("_mapped_values_foo", "_mapped_values_foo"),
        ("_mapped_values_foo_mapped_values", "_mapped_values_foo"),
    ],
)
def test_get_original_col_name(col_name: str, expected: str) -> None:
    assert get_original_col_name(col_name) == expected



from datetime import datetime

import pytest

from tests.helpers import ColumnCreator
from vivarium.framework.randomness import RandomnessStream
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.resource import Resource
from vivarium.framework.resource.resource import Column, NullResource
from vivarium.framework.values import Pipeline, ValueModifier, ValueSource


def test_resource_id() -> None:
    resource = Resource("value_source", "test", ColumnCreator())
    assert resource.resource_id == "value_source.test"


@pytest.mark.parametrize(
    "resource, is_initialized",
    [
        (Pipeline("foo"), False),
        (ValueSource(Pipeline("bar"), lambda: 1, ColumnCreator()), False),
        (ValueModifier(Pipeline("baz"), lambda: 1, ColumnCreator()), False),
        (Column("foo", ColumnCreator()), True),
        (RandomnessStream("bar", lambda: datetime.now(), 1, IndexMap()), False),
        (NullResource(0, ColumnCreator()), True),
    ],
)
def test_resource_is_initialized(resource: Resource, is_initialized: bool) -> None:
    assert resource.is_initialized == is_initialized



"""
======================
Results System Manager
======================

"""

from __future__ import annotations

from collections import defaultdict
from enum import Enum
from typing import TYPE_CHECKING, Any, Iterable, Sequence

import pandas as pd

from vivarium.framework.event import Event
from vivarium.framework.results.context import ResultsContext
from vivarium.framework.results.observation import Observation
from vivarium.framework.values import Pipeline
from vivarium.manager import Manager
from vivarium.types import ScalarMapper, VectorMapper

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder


class SourceType(Enum):
    COLUMN = 0
    VALUE = 1


class ResultsManager(Manager):
    """Backend manager object for the results management system.

    This class contains the public methods used by the :class:`ResultsInterface <vivarium.framework.results.interface.ResultsInterface>`
    to register stratifications and observations as well as the :meth:`get_results <get_results>`
    method used to retrieve formatted results by the :class:`ResultsContext <vivarium.framework.results.context.ResultsContext>`.

    """

    CONFIGURATION_DEFAULTS = {
        "stratification": {
            "default": [],
            "excluded_categories": {},
        }
    }

    def __init__(self) -> None:
        self._raw_results: defaultdict[str, pd.DataFrame] = defaultdict()
        self._results_context = ResultsContext()
        self._required_columns = {"tracked"}
        self._required_values: set[Pipeline] = set()
        self._name = "results_manager"

    @property
    def name(self) -> str:
        return self._name

    def get_results(self) -> dict[str, pd.DataFrame]:
        """Return the measure-specific formatted results in a dictionary.

        Notes
        -----
        self._results_context.observations is a list where each item is a dictionary
        of the form {lifecycle_phase: {(pop_filter, stratification_names): List[Observation]}}.
        We use a triple-nested for loop to iterate over only the list of Observations
        (i.e. we do not need the lifecycle_phase, pop_filter, or stratification_names
        for this method).

        Returns
        -------
            A dictionary of measure-specific formatted results. The keys are the
            measure names and the values are the respective results.
        """
        formatted = {}
        for observation_details in self._results_context.observations.values():
            for observations in observation_details.values():
                for observation in observations:
                    measure = observation.name
                    results = self._raw_results[measure].copy()
                    formatted[measure] = observation.results_formatter(measure, results)
        return formatted

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: "Builder") -> None:
        """Set up the results manager."""
        self._results_context.setup(builder)

        self.logger = builder.logging.get_logger(self.name)
        self.population_view = builder.population.get_view([])
        self.clock = builder.time.clock()
        self.step_size = builder.time.step_size()

        builder.event.register_listener("post_setup", self.on_post_setup)
        builder.event.register_listener("time_step__prepare", self.on_time_step_prepare)
        builder.event.register_listener("time_step", self.on_time_step)
        builder.event.register_listener("time_step__cleanup", self.on_time_step_cleanup)
        builder.event.register_listener("collect_metrics", self.on_collect_metrics)

        self.get_value = builder.value.get_value

        self.set_default_stratifications(builder)

    def on_post_setup(self, _: Event) -> None:
        """Initialize results for each measure."""
        registered_stratifications = self._results_context.stratifications

        used_stratifications = set()
        for lifecycle_phase in self._results_context.observations:
            for (
                _pop_filter,
                event_requested_stratification_names,
            ), observations in self._results_context.observations[lifecycle_phase].items():
                if event_requested_stratification_names is not None:
                    used_stratifications |= set(event_requested_stratification_names)
                for observation in observations:
                    measure = observation.name
                    self._raw_results[measure] = observation.results_initializer(
                        event_requested_stratification_names, registered_stratifications
                    )

        registered_stratification_names = set(
            stratification.name for stratification in registered_stratifications
        )
        unused_stratifications = registered_stratification_names - used_stratifications
        if unused_stratifications:
            self.logger.info(
                "The following stratifications are registered but not used by any "
                f"observers: \n{sorted(list(unused_stratifications))}"
            )
        missing_stratifications = used_stratifications - registered_stratification_names
        if missing_stratifications:
            raise ValueError(
                "The following observers are requested to be stratified by "
                f"stratifications that are not registered: \n{sorted(list(missing_stratifications))}"
            )

    def on_time_step_prepare(self, event: Event) -> None:
        """Define the listener callable for the time_step__prepare phase."""
        self.gather_results("time_step__prepare", event)

    def on_time_step(self, event: Event) -> None:
        """Define the listener callable for the time_step phase."""
        self.gather_results("time_step", event)

    def on_time_step_cleanup(self, event: Event) -> None:
        """Define the listener callable for the time_step__cleanup phase."""
        self.gather_results("time_step__cleanup", event)

    def on_collect_metrics(self, event: Event) -> None:
        """Define the listener callable for the collect_metrics phase."""
        self.gather_results("collect_metrics", event)

    def gather_results(self, lifecycle_phase: str, event: Event) -> None:
        """Update existing results with any new results."""
        population = self._prepare_population(event)
        if population.empty:
            return
        for results_group, measure, updater in self._results_context.gather_results(
            population, lifecycle_phase, event
        ):
            if results_group is not None and measure is not None and updater is not None:
                self._raw_results[measure] = updater(
                    self._raw_results[measure], results_group
                )

    ##########################
    # Stratification methods #
    ##########################

    def set_default_stratifications(self, builder: "Builder") -> None:
        """Set the default stratifications for the results context.

        This passes the default stratifications from the configuration to the
        :class:`ResultsContext <vivarium.framework.results.context.ResultsContext>`
        :meth:`set_default_stratifications` method to be set.

        Parameters
        ----------
        builder
            The builder object for the simulation.
        """
        default_stratifications = builder.configuration.stratification.default
        self._results_context.set_default_stratifications(default_stratifications)

    def register_stratification(
        self,
        name: str,
        categories: list[str],
        excluded_categories: list[str] | None,
        mapper: VectorMapper | ScalarMapper | None,
        is_vectorized: bool,
        requires_columns: list[str] = [],
        requires_values: list[str] = [],
    ) -> None:
        """Manager-level stratification registration.

        Adds a stratification to the
        :class:`ResultsContext <vivarium.framework.results.context.ResultsContext>`
        as well as the stratification's required resources to this manager.

        Parameters
        ----------
        name
            Name of the stratification.
        categories
            Exhaustive list of all possible stratification values.
        excluded_categories
            List of possible stratification values to exclude from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        mapper
            A callable that maps the columns and value pipelines specified by the
            `requires_columns` and `requires_values` arguments to the stratification
            categories. It can either map the entire population or an individual
            simulant. A simulation will fail if the `mapper` ever produces an invalid
            value.
        is_vectorized
            True if the `mapper` function will map the entire population, and False
            if it will only map a single simulant.
        requires_columns
            A list of the state table columns that are required by the `mapper`
            to produce the stratification.
        requires_values
            A list of the value pipelines that are required by the `mapper` to
            produce the stratification.
        """
        self.logger.debug(f"Registering stratification {name}")
        target_columns = list(requires_columns) + list(requires_values)
        self._results_context.add_stratification(
            name, target_columns, categories, excluded_categories, mapper, is_vectorized
        )
        self._add_resources(requires_columns, SourceType.COLUMN)
        self._add_resources(requires_values, SourceType.VALUE)

    def register_binned_stratification(
        self,
        target: str,
        binned_column: str,
        bin_edges: Sequence[int | float],
        labels: list[str],
        excluded_categories: list[str] | None,
        target_type: str,
        **cut_kwargs: int | str | bool,
    ) -> None:
        """Manager-level registration of a continuous `target` quantity to observe
        into bins in a `binned_column`.

        Parameters
        ----------
        target
            Name of the state table column or value pipeline to be binned.
        binned_column
            Name of the (binned) stratification.
        bin_edges
            List of scalars defining the bin edges, passed to :meth: pandas.cut.
            The length must be equal to the length of `labels` plus 1.
        labels
            List of string labels for bins. The length must be equal to the length
            of `bin_edges` minus 1.
        excluded_categories
            List of possible stratification values to exclude from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        target_type
            Type specification of the `target` to be binned. "column" if it's a
            state table column or "value" if it's a value pipeline.
        **cut_kwargs
            Keyword arguments for :meth: pandas.cut.
        """
        if not isinstance(labels, list) or not all(
            [isinstance(label, str) for label in labels]
        ):
            raise ValueError(
                f"Labels must be a list of strings when registering a binned stratification, but labels was {labels} when registering the {binned_column} stratification."
            )

        def _bin_data(data: pd.DataFrame) -> pd.Series[Any]:
            """Use pandas.cut to bin continuous values"""
            data = data.squeeze(axis=1)
            if not isinstance(data, pd.Series):
                raise ValueError(f"Expected a Series, but got type {type(data)}.")
            data = pd.cut(
                data, bin_edges, labels=labels, right=False, include_lowest=True, **cut_kwargs
            )
            return data

        if len(bin_edges) != len(labels) + 1:
            raise ValueError(
                f"The number of bin edges plus 1 ({len(bin_edges)+1}) does not "
                f"match the number of labels ({len(labels)})"
            )

        target_arg = "requires_columns" if target_type == "column" else "requires_values"
        target_kwargs = {target_arg: [target]}

        self.register_stratification(
            name=binned_column,
            categories=labels,
            excluded_categories=excluded_categories,
            mapper=_bin_data,
            is_vectorized=True,
            **target_kwargs,
        )

    def register_observation(
        self,
        observation_type: type[Observation],
        is_stratified: bool,
        name: str,
        pop_filter: str,
        when: str,
        requires_columns: list[str],
        requires_values: list[str],
        **kwargs: Any,
    ) -> None:
        """Manager-level observation registration.

        Adds an observation to the
        :class:`ResultsContext <vivarium.framework.results.context.ResultsContext>`
        as well as the observation's required resources to this manager.

        Parameters
        ----------
        observation_type
            Specific class type of observation to register.
        is_stratified
            True if the observation is a stratified type and False if not.
        name
            Name of the observation. It will also be the name of the output results file
            for this particular observation.
        pop_filter
            A Pandas query filter string to filter the population down to the simulants who should
            be considered for the observation.
        when
            Name of the lifecycle phase the observation should happen. Valid values are:
            "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
        requires_columns
            List of the state table columns that are required by either the `pop_filter` or the `aggregator`.
        requires_values
            List of the value pipelines that are required by either the `pop_filter` or the `aggregator`.
        **kwargs
            Additional keyword arguments to be passed to the observation's constructor.
        """
        self.logger.debug(f"Registering observation {name}")

        if is_stratified:
            # Resolve required stratifications and add to kwargs dictionary
            additional_stratifications = kwargs.get("additional_stratifications", [])
            excluded_stratifications = kwargs.get("excluded_stratifications", [])
            self._warn_check_stratifications(
                additional_stratifications, excluded_stratifications
            )
            stratifications = self._get_stratifications(
                list(kwargs.get("stratifications", [])),
                list(additional_stratifications),
                list(excluded_stratifications),
            )
            kwargs["stratifications"] = stratifications
            # Remove the unused kwargs before passing to the results context registration
            del kwargs["additional_stratifications"]
            del kwargs["excluded_stratifications"]

        self._add_resources(requires_columns, SourceType.COLUMN)
        self._add_resources(requires_values, SourceType.VALUE)

        self._results_context.register_observation(
            observation_type=observation_type,
            name=name,
            pop_filter=pop_filter,
            when=when,
            **kwargs,
        )

    ##################
    # Helper methods #
    ##################

    def _get_stratifications(
        self,
        stratifications: list[str] = [],
        additional_stratifications: list[str] = [],
        excluded_stratifications: list[str] = [],
    ) -> tuple[str, ...]:
        """Resolve the stratifications required for the observation."""
        stratifications = list(
            set(
                self._results_context.default_stratifications
                + stratifications
                + additional_stratifications
            )
            - set(excluded_stratifications)
        )
        # Makes sure measure identifiers have fields in the same relative order.
        return tuple(sorted(stratifications))

    def _add_resources(self, target: list[str], target_type: SourceType) -> None:
        """Add required resources to the manager's list of required columns and values."""
        if len(target) == 0:
            return  # do nothing on empty lists
        target_set = set(target) - {"event_time", "current_time", "event_step_size"}
        if target_type == SourceType.COLUMN:
            self._required_columns.update(target_set)
        elif target_type == SourceType.VALUE:
            self._required_values.update([self.get_value(target) for target in target_set])

    def _prepare_population(self, event: Event) -> pd.DataFrame:
        """Prepare the population for results gathering."""
        population = self.population_view.subview(list(self._required_columns)).get(
            event.index
        )
        population["current_time"] = self.clock()
        population["event_step_size"] = event.step_size
        population["event_time"] = self.clock() + event.step_size  # type: ignore [operator]
        for k, v in event.user_data.items():
            population[k] = v
        for pipeline in self._required_values:
            population[pipeline.name] = pipeline(event.index)
        return population

    def _warn_check_stratifications(
        self, additional_stratifications: list[str], excluded_stratifications: list[str]
    ) -> None:
        """Check additional and excluded stratifications if they'd not affect
        stratifications (i.e., would be NOP), and emit warning."""
        nop_additional = [
            s
            for s in additional_stratifications
            if s in self._results_context.default_stratifications
        ]
        if len(nop_additional):
            self.logger.warning(
                f"Specified additional stratifications are already included by default: {nop_additional}",
            )
        nop_exclude = [
            s
            for s in excluded_stratifications
            if s not in self._results_context.default_stratifications
        ]
        if len(nop_exclude):
            self.logger.warning(
                f"Specified excluded stratifications are already not included by default: {nop_exclude}",
            )



"""
====================
The Artifact Manager
====================

This module contains the :class:`ArtifactManager`, a ``vivarium`` plugin
for handling complex data bound up in a data artifact.

"""
from __future__ import annotations

import re
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pandas as pd
from layered_config_tree.main import LayeredConfigTree

from vivarium.framework.artifact import ArtifactException
from vivarium.framework.artifact.artifact import Artifact
from vivarium.manager import Interface, Manager

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder
    from vivarium.types import ScalarValue


class ArtifactManager(Manager):
    """The controller plugin component for managing a data artifact."""

    CONFIGURATION_DEFAULTS = {
        "input_data": {
            "artifact_path": None,
            "artifact_filter_term": None,
            "input_draw_number": None,
        }
    }

    def __init__(self) -> None:
        self._default_value_column = "value"

    @property
    def name(self) -> str:
        return "artifact_manager"

    def setup(self, builder: Builder) -> None:
        """Performs this component's simulation setup."""
        self.logger = builder.logging.get_logger(self.name)
        # because not all columns are accessible via artifact filter terms, apply config filters separately
        self.config_filter_term = validate_filter_term(
            builder.configuration.input_data.artifact_filter_term
        )
        self.artifact = self._load_artifact(builder.configuration)
        builder.lifecycle.add_constraint(self.load, allow_during=["setup"])

    def _load_artifact(self, configuration: LayeredConfigTree) -> Artifact | None:
        """Loads artifact data.

        Looks up the path to the artifact hdf file, builds a default filter,
        and generates the data artifact. Stores any configuration specified filter
        terms separately to be applied on loading, because not all columns are
        available via artifact filter terms.

        Parameters
        ----------
        configuration
            Configuration block of the model specification containing the input data parameters.

        Returns
        -------
            An interface to the data artifact.
        """
        if not configuration.input_data.artifact_path:
            return None
        artifact_path = parse_artifact_path_config(configuration)
        base_filter_terms = get_base_filter_terms(configuration)
        self.logger.info(f"Running simulation from artifact located at {artifact_path}.")
        self.logger.info(f"Artifact base filter terms are {base_filter_terms}.")
        self.logger.info(f"Artifact additional filter terms are {self.config_filter_term}.")
        return Artifact(artifact_path, base_filter_terms)

    def load(self, entity_key: str, **column_filters: int | str | Sequence[int | str]) -> Any:
        """Loads data associated with the given entity key.

        Parameters
        ----------
        entity_key
            The key associated with the expected data.
        column_filters
            Filters that subset the data by a categorical column and then remove
            the column from the raw data. They are supplied as keyword arguments
            to the load method in the form "column=value".

        Returns
        -------
            The data associated with the given key, filtered down to the
            requested subset if the data is a dataframe.
        """
        if self.artifact is None:
            raise ArtifactException("No artifact defined for simulation.")

        data = self.artifact.load(entity_key)
        if isinstance(data, pd.DataFrame):  # could be metadata dict
            data = data.reset_index()
            draw_col = [c for c in data.columns if "draw" in c]
            if draw_col:
                data = data.rename(columns={draw_col[0]: self._default_value_column})

            data = filter_data(data, self.config_filter_term, **column_filters)

        return data

    def value_columns(
        self,
    ) -> Callable[[str | pd.DataFrame | dict[str, list[ScalarValue] | list[str]]], list[str]]:
        """Returns a function that returns the value columns for the given input.

        The function can be called with either a string or a pandas DataFrame.
        If a string is provided, it is interpreted as an artifact key, and the
        value columns for the data stored at that key are returned.

        Currently, the returned function will always return ["value"].

        Returns
        -------
            A function that returns the value columns for the given input.
        """
        return lambda _: [self._default_value_column]

    def __repr__(self) -> str:
        return "ArtifactManager()"


class ArtifactInterface(Interface):
    """The builder interface for accessing a data artifact."""

    def __init__(self, manager: ArtifactManager) -> None:
        self._manager = manager

    def load(self, entity_key: str, **column_filters: int | str | Sequence[int | str]) -> Any:
        """Loads data associated with a formatted entity key.

        The provided entity key must be of the form
        {entity_type}.{measure} or {entity_type}.{entity_name}.{measure}.

        Here entity_type denotes the kind of entity being described. Examples
        include cause, risk, population, and covariates.

        The entity_name is the name of the specific entity. For example,
        if we had entity_type as cause, we might have entity_name as
        diarrheal_diseases or ischemic_heart_disease.

        Finally, measure is the name of the quantity the data describes.
        Examples of measures are incidence, disability_weight, relative_risk,
        and cost.

        Parameters
        ----------
        entity_key
            The key associated with the expected data.
        column_filters
            Filters that subset the data by a categorical column and then
            remove the column from the raw data. They are supplied as keyword
            arguments to the load method in the form "column=value".

        Returns
        -------
            The data associated with the given key filtered down to the requested subset.
        """
        return self._manager.load(entity_key, **column_filters)

    def value_columns(
        self,
    ) -> Callable[[str | pd.DataFrame | dict[str, list[ScalarValue] | list[str]]], list[str]]:
        """Returns a function that returns the value columns for the given input.

        The function can be called with either a string or a pandas DataFrame.
        If a string is provided, it is interpreted as an artifact key, and the
        value columns for the data stored at that key are returned.

        Returns
        -------
            A function that returns the value columns for the given input.
        """
        return self._manager.value_columns()

    def __repr__(self) -> str:
        return "ArtifactManagerInterface()"


def filter_data(
    data: pd.DataFrame,
    config_filter_term: str | None = None,
    **column_filters: int | str | Sequence[int | str],
) -> pd.DataFrame:
    """Uses the provided column filters and age_group conditions to subset the raw data."""
    data = _config_filter(data, config_filter_term)
    data = _subset_rows(data, **column_filters)
    data = _subset_columns(data, **column_filters)
    return data


def _config_filter(data: pd.DataFrame, config_filter_term: str | None) -> pd.DataFrame:
    if config_filter_term:
        filter_column = re.split("[<=>]", config_filter_term.split()[0])[0]
        if filter_column in data.columns:
            data = data.query(config_filter_term)
    return data


def validate_filter_term(config_filter_term: str | None) -> str | None:
    multiple_filter_indicators = [" and ", " or ", "|", "&"]
    if config_filter_term is not None and any(
        x in config_filter_term for x in multiple_filter_indicators
    ):
        raise NotImplementedError(
            "Only a single filter term via the configuration is currently supported."
        )
    return config_filter_term


def _subset_rows(
    data: pd.DataFrame, **column_filters: int | str | Sequence[int | str]
) -> pd.DataFrame:
    """Filters out unwanted rows from the data using the provided filters."""
    extra_filters = set(column_filters.keys()) - set(data.columns)
    if extra_filters:
        raise ValueError(
            f"Filtering by non-existent columns: {extra_filters}. "
            f"Available columns: {data.columns}"
        )

    for column, condition in column_filters.items():
        if isinstance(condition, (str, int)):
            condition = [condition]
        mask = pd.Series(False, index=data.index)
        for c in condition:
            mask |= data[f"{column}"] == c
        row_indexer = data[mask].index
        data = data.loc[row_indexer, :]

    return data


def _subset_columns(
    data: pd.DataFrame, **column_filters: int | str | Sequence[int | str]
) -> pd.DataFrame:
    """Filters out unwanted columns and default columns from the data using provided filters."""
    columns_to_remove = set(list(column_filters.keys()) + ["draw"])
    columns_to_remove = columns_to_remove.intersection(data.columns)
    return data.drop(columns=list(columns_to_remove))


def get_base_filter_terms(configuration: LayeredConfigTree) -> list[str]:
    """Parses default filter terms from the artifact configuration."""
    base_filter_terms = []

    draw = configuration.input_data.input_draw_number
    if draw is not None:
        base_filter_terms.append(f"draw == {draw}")

    return base_filter_terms


def parse_artifact_path_config(config: LayeredConfigTree) -> str:
    """Gets the path to the data artifact from the simulation configuration.

    The path specified in the configuration may be absolute or it may be relative
    to the location of the configuration file.

    Parameters
    ----------
    config
        The configuration block of the simulation model specification containing the artifact path.

    Returns
    -------
        The path to the data artifact.
    """
    path = Path(config.input_data.artifact_path)

    if not path.is_absolute():
        path_config = config.input_data.metadata("artifact_path")[-1]
        if path_config["source"] is None:
            raise ValueError("Insufficient information provided to find artifact.")
        path = Path(path_config["source"]).parent.joinpath(path).resolve()

    if not path.exists():
        raise FileNotFoundError(f"Cannot find artifact at path {path}")

    return str(path)



import re
from types import MethodType

import numpy as np
import pandas as pd
import pytest
import pytest_mock
from _pytest.logging import LogCaptureFixture
from layered_config_tree.main import LayeredConfigTree
from loguru import logger
from pandas.api.types import CategoricalDtype

from tests.framework.results.helpers import (
    BIN_BINNED_COLUMN,
    BIN_LABELS,
    BIN_SILLY_BIN_EDGES,
    BIN_SOURCE,
    FAMILIARS,
    HARRY_POTTER_CONFIG,
    HOUSE_CATEGORIES,
    NAME,
    NAME_COLUMNS,
    POWER_LEVEL_GROUP_LABELS,
    STUDENT_HOUSES,
    CatBombObserver,
    ExamScoreObserver,
    FullyFilteredHousePointsObserver,
    Hogwarts,
    HogwartsResultsStratifier,
    HousePointsObserver,
    MagicalAttributesObserver,
    NoStratificationsQuidditchWinsObserver,
    QuidditchWinsObserver,
    ValedictorianObserver,
    mock_get_value,
    sorting_hat_serial,
    sorting_hat_vectorized,
    verify_stratification_added,
)
from vivarium.framework.results import VALUE_COLUMN
from vivarium.framework.results.context import ResultsContext
from vivarium.framework.results.manager import ResultsManager
from vivarium.framework.results.observation import AddingObservation
from vivarium.interface.interactive import InteractiveContext
from vivarium.types import ScalarMapper, VectorMapper


@pytest.mark.parametrize(
    "stratifications, default_stratifications, additional_stratifications, excluded_stratifications, expected_stratifications",
    [
        ([], [], [], [], []),
        (
            [],
            ["age", "sex"],
            ["handedness"],
            ["age"],
            ["sex", "handedness"],
        ),
        ([], ["age", "sex"], [], ["age", "sex"], []),
        ([], ["age"], [], ["bogus_exclude_column"], ["age"]),
        (["custom"], ["age", "sex"], [], [], ["custom", "age", "sex"]),
    ],
    ids=[
        "empty_add_empty_exclude",
        "one_add_one_exclude",
        "all_defaults_excluded",
        "bogus_exclude",
        "custom_stratification",
    ],
)
def test__get_stratifications(
    stratifications: list[str],
    default_stratifications: list[str],
    additional_stratifications: list[str],
    excluded_stratifications: list[str],
    expected_stratifications: list[str],
    mocker: pytest_mock.MockFixture,
) -> None:
    ctx = ResultsContext()
    ctx.default_stratifications = default_stratifications
    mgr = ResultsManager()
    mocker.patch.object(mgr, "_results_context", ctx)
    # default_stratifications would normally be set via ResultsInterface.set_default_stratifications()
    final_stratifications = mgr._get_stratifications(
        stratifications, additional_stratifications, excluded_stratifications
    )
    assert sorted(final_stratifications) == sorted(expected_stratifications)


#######################################
# Tests for `register_stratification` #
#######################################


@pytest.mark.parametrize(
    "name, sources, categories, excluded_categories, mapper, is_vectorized",
    [
        (
            NAME,
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            [],
            sorting_hat_vectorized,
            True,
        ),
        (
            NAME,
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            [],
            sorting_hat_serial,
            False,
        ),
        (
            NAME,
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            ["gryffindor"],
            sorting_hat_vectorized,
            True,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper", "excluded_categories"],
)
def test_register_stratification_no_pipelines(
    name: str,
    sources: list[str],
    categories: list[str],
    excluded_categories: list[str],
    mapper: VectorMapper | ScalarMapper,
    is_vectorized: bool,
    mocker: pytest_mock.MockFixture,
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {}}
    )
    mgr.setup(builder)
    mgr.register_stratification(
        name=name,
        categories=categories,
        excluded_categories=excluded_categories,
        mapper=mapper,
        is_vectorized=is_vectorized,
        requires_columns=sources,
    )
    for item in sources:
        assert item in mgr._required_columns
    assert verify_stratification_added(
        mgr._results_context.stratifications,
        name,
        sources,
        categories,
        excluded_categories,
        mapper,
        is_vectorized,
    )


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized",
    [
        (
            NAME,
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            sorting_hat_vectorized,
            True,
        ),
        (
            NAME,
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            sorting_hat_serial,
            False,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_register_stratification_with_pipelines(
    name: str,
    sources: list[str],
    categories: list[str],
    mapper: VectorMapper | ScalarMapper,
    is_vectorized: bool,
    mocker: pytest_mock.MockFixture,
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {}}
    )
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    mgr.register_stratification(
        name=name,
        categories=categories,
        excluded_categories=None,
        mapper=mapper,
        is_vectorized=is_vectorized,
        requires_columns=[],
        requires_values=sources,
    )
    for item in sources:
        assert item in [pipeline.name for pipeline in mgr._required_values]
    assert verify_stratification_added(
        mgr._results_context.stratifications,
        name,
        sources,
        categories,
        [],
        mapper,
        is_vectorized,
    )


@pytest.mark.parametrize(
    "name, sources, categories, mapper, is_vectorized",
    [
        (  # expected Stratification for vectorized
            NAME,
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            sorting_hat_vectorized,
            True,
        ),
        (  # expected Stratification for non-vectorized
            NAME,
            NAME_COLUMNS,
            HOUSE_CATEGORIES,
            sorting_hat_serial,
            False,
        ),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_register_stratification_with_column_and_pipelines(
    name: str,
    sources: list[str],
    categories: list[str],
    mapper: VectorMapper | ScalarMapper,
    is_vectorized: bool,
    mocker: pytest_mock.MockFixture,
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {}}
    )
    # Set up mock builder with mocked get_value call for Pipelines
    mocker.patch.object(builder, "value.get_value")
    builder.value.get_value = MethodType(mock_get_value, builder)
    mgr.setup(builder)
    mocked_column_name = "silly_column"
    mgr.register_stratification(
        name=name,
        categories=categories,
        excluded_categories=None,
        mapper=mapper,
        is_vectorized=is_vectorized,
        requires_columns=[mocked_column_name],
        requires_values=sources,
    )
    assert mocked_column_name in mgr._required_columns
    for item in sources:
        assert item in [pipeline.name for pipeline in mgr._required_values]
    all_sources = sources.copy()
    all_sources.append(mocked_column_name)
    assert verify_stratification_added(
        mgr._results_context.stratifications,
        name,
        all_sources,
        categories,
        [],
        mapper,
        is_vectorized,
    )


##############################################
# Tests for `register_binned_stratification` #
##############################################


@pytest.mark.parametrize(
    "bins, labels",
    [(BIN_SILLY_BIN_EDGES, BIN_LABELS[1:]), (BIN_SILLY_BIN_EDGES[1:], BIN_LABELS)],
    ids=["too_many_bins", "too_many_labels"],
)
def test_register_binned_stratification_raises_bins_labels_mismatch(
    bins: list[float], labels: list[str]
) -> None:
    mgr = ResultsManager()
    with pytest.raises(
        ValueError,
        match=r"The number of bin edges plus 1 \(\d+\) does not match the number of labels \(\d+\)",
    ):
        mgr.register_binned_stratification(
            target=BIN_SOURCE,
            binned_column=BIN_BINNED_COLUMN,
            bin_edges=bins,
            labels=labels,
            excluded_categories=None,
            target_type="column",
        )


from vivarium.types import VectorMapper


def test_binned_stratification_mapper() -> None:
    mgr = ResultsManager()
    mgr.logger = logger
    mgr.register_binned_stratification(
        target=BIN_SOURCE,
        binned_column=BIN_BINNED_COLUMN,
        bin_edges=BIN_SILLY_BIN_EDGES,
        labels=BIN_LABELS,
        excluded_categories=None,
        target_type="column",
    )
    strat = mgr._results_context.stratifications[0]
    data = pd.DataFrame([-np.inf] + BIN_SILLY_BIN_EDGES + [np.inf])
    groups = strat.vectorized_mapper(data)
    expected_groups = pd.Series([np.nan] + BIN_LABELS + [np.nan] * 2)
    assert (groups.isna() == expected_groups.isna()).all()
    assert (groups[groups.notna()] == expected_groups[expected_groups.notna()]).all()


@pytest.mark.parametrize(
    "default, additional, excluded, match",
    [
        (["age", "sex"], ["age"], [], ["age"]),
        (["age", "sex"], [], ["eye_color"], ["eye_color"]),
        (["age", "sex"], ["age"], ["eye_color"], ["age", "eye_color"]),
    ],
    ids=[
        "additional_no_operation",
        "exclude_no_operation",
        "additional_and_exclude_no_operation",
    ],
)
def test_add_observation_nop_stratifications(
    default: list[str],
    additional: list[str],
    excluded: list[str],
    match: list[str],
    mocker: pytest_mock.MockFixture,
    caplog: LogCaptureFixture,
) -> None:
    mgr = ResultsManager()
    builder = mocker.Mock()
    mgr.setup(builder)
    mgr.logger = logger

    mgr._results_context.default_stratifications = default
    mgr.register_observation(
        observation_type=AddingObservation,
        is_stratified=True,
        name="name",
        pop_filter='alive == "alive"',
        aggregator_sources=[],
        aggregator=lambda: None,
        requires_columns=[],
        requires_values=[],
        additional_stratifications=additional,
        excluded_stratifications=excluded,
        when="collect_metrics",
        results_formatter=lambda: None,
    )
    for m in match:
        assert m in caplog.text


def test_setting_default_stratifications_at_setup(mocker: pytest_mock.MockFixture) -> None:
    """Test that set default stratifications happens at setup"""
    mgr = ResultsManager()
    builder = mocker.Mock()
    mocker.patch.object(mgr._results_context, "set_default_stratifications")
    mgr._results_context.set_default_stratifications.assert_not_called()  # type: ignore[attr-defined]

    mgr.setup(builder)

    mgr._results_context.set_default_stratifications.assert_called_once_with(  # type: ignore[attr-defined]
        builder.configuration.stratification.default
    )


################################
# Tests for results processing #
################################


def test__raw_results_initialized_as_empty_dict(mocker: pytest_mock.MockFixture) -> None:
    """Test that raw results are initialized as an empty dictionary"""
    mgr = ResultsManager()
    builder = mocker.Mock()
    mgr.setup(builder)
    assert mgr._raw_results == {}


def test_stratified__raw_results_initialization() -> None:
    """Test that raw results are being initialized correctly. We expect a dictionary
    of pd.DataFrames. Each key of the dictionary is an observed measure name and
    the corresponding value is a zeroed-out multiindex pd.DataFrame of that observer's
    stratifications.
    """

    components = [
        HousePointsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]

    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    raw_results = sim._results._raw_results
    assert isinstance(raw_results, dict)
    assert set(raw_results) == set(["house_points", "quidditch_wins"])
    for measure in raw_results:
        result = raw_results[measure]
        assert isinstance(result, pd.DataFrame)
        assert (result[VALUE_COLUMN] == 0).all()
    STUDENT_HOUSES_LIST = list(STUDENT_HOUSES)

    # Check that indexes are as expected

    # Multi-stratification index is type MultiIndex where each layer dtype is Category
    expected_house_points_multi_idx = pd.MultiIndex.from_product(
        [STUDENT_HOUSES_LIST, POWER_LEVEL_GROUP_LABELS],
        names=["student_house", "power_level"],
    )
    # HACK: We need to set the levels to be CategoricalDtype but you can't set that
    # directly on the MultiIndex. Convert to df, set type, convert back
    expected_house_points_idx = (
        pd.DataFrame(index=expected_house_points_multi_idx)
        .reset_index()
        .astype(CategoricalDtype)
        .set_index(["student_house", "power_level"])
        .index
    )
    assert raw_results["house_points"].index.equals(expected_house_points_idx)

    # Single-stratification index is type CategoricalIndex
    expected_quidditch_idx = pd.CategoricalIndex(FAMILIARS, name="familiar")
    assert raw_results["quidditch_wins"].index.equals(expected_quidditch_idx)


def test_no_stratifications__raw_results_initialization() -> None:
    """Test that Observers requesting no stratifications result in a
    single-row DataFrame with 'value' of zero and index labeled 'all'
    """
    components = [Hogwarts(), NoStratificationsQuidditchWinsObserver()]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    raw_results = sim._results._raw_results["no_stratifications_quidditch_wins"]
    assert isinstance(raw_results, pd.DataFrame)
    assert raw_results.shape == (1, 1)
    assert raw_results[VALUE_COLUMN].iat[0] == 0
    assert raw_results.index.equals(pd.Index(["all"]))


def test_observers_with_missing_stratifications_fail() -> None:
    """Test that an error is raised if an Observer requests a stratification
    that never actually gets registered.
    """
    components = [QuidditchWinsObserver(), HousePointsObserver(), Hogwarts()]

    expected_missing = ["familiar", "power_level_group", "student_house"]
    expected_log_msg = re.escape(
        "The following observers are requested to be stratified by stratifications "
        f"that are not registered: \n{expected_missing}"
    )

    with pytest.raises(ValueError, match=expected_log_msg):
        InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)


def test_unused_stratifications_are_logged(caplog: LogCaptureFixture) -> None:
    """Test that we issue a logger.info warning if Stratifications are registered
    but never actually used by an Observer

    The HogwartsResultsStratifier registers "student_house", "familiar", and
    "power_level_group" stratifiers. However, we will only use the QuidditchWinsObserver
    which only uses "familiar" and the MagicalAttributesObserver which only uses
    "power_level_group". We would thus expect only "student_house" to be logged
    as an unused stratification.

    """
    components = [
        Hogwarts(),
        HogwartsResultsStratifier(),
        QuidditchWinsObserver(),
        MagicalAttributesObserver(),
    ]
    InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)

    log_split = caplog.text.split(
        "The following stratifications are registered but not used by any observers: \n"
    )
    # Check that the log message is present and only exists one time
    assert len(log_split) == 2
    # Check that the log message contains the expected Stratifications
    assert "['student_house']" in log_split[1]


def test_stratified_observation_results() -> None:
    components = [
        Hogwarts(),
        CatBombObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    assert (sim.get_results()["cat_bomb"]["value"] == 0.0).all()
    sim.step()
    num_familiars = sim.get_population().groupby(["familiar", "student_house"]).apply(len)
    expected = num_familiars.loc["cat"] ** 1.0
    expected.name = "value"
    assert expected.sort_values().equals(sim.get_results()["cat_bomb"]["value"].sort_values())
    sim.step()
    num_familiars = sim.get_population().groupby(["familiar", "student_house"]).apply(len)
    expected = num_familiars.loc["cat"] ** 2.0
    expected.name = "value"
    assert expected.sort_values().equals(sim.get_results()["cat_bomb"]["value"].sort_values())


def test_unstratified_observation_results() -> None:
    components = [
        Hogwarts(),
        ValedictorianObserver(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    first_valedictorian = sim.get_results()["valedictorian"]
    assert len(first_valedictorian) == 1
    sim.step()
    second_valedictorian = sim.get_results()["valedictorian"]
    assert len(second_valedictorian) == 1
    assert (
        first_valedictorian["student_id"].iat[0] != second_valedictorian["student_id"].iat[0]
    )


def test_concatenating_observation_results() -> None:
    components = [
        Hogwarts(),
        ExamScoreObserver(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    results_one_step = sim.get_results()["exam_score"]
    assert (results_one_step["exam_score"] == 10.0).all()
    sim.step()
    expected_two_steps = results_one_step.copy()
    expected_two_steps["exam_score"] = 20.0
    expected_two_steps["event_time"] = results_one_step["event_time"] + pd.Timedelta(
        days=sim.configuration.time.step_size
    )
    assert sim.get_results()["exam_score"].equals(
        pd.concat([results_one_step, expected_two_steps], axis=0).reset_index(drop=True)
    )


def test_adding_observation_results() -> None:
    """Test that adding observation results are being updated correctly."""

    def _check_house_points(pop: pd.DataFrame, step_number: int) -> None:
        """We know that house points are stratified by 'student_house' and 'power_level_group'.
        and that each wizard of gryffindor and of level 20 or 80 (which correspond
        to 'low' and 'very high' power level groups) gains a point
        """
        assert set(pop["house_points"]) == set([0, 1])
        assert (pop.loc[pop["house_points"] != 0, "student_house"] == "gryffindor").all()
        assert set(pop.loc[pop["house_points"] != 0, "power_level"]) == set([20, 80])
        group_sizes = pd.DataFrame(
            pop.groupby(["student_house", "power_level"]).size().astype("float"),
            columns=[VALUE_COLUMN],
        )
        raw_results = sim._results._raw_results["house_points"]
        # We cannot use `equals` here because raw results have a MultiIndex where
        # each layer is a Category dtype but pop has object dtype for the relevant columns
        assert (  # type: ignore[union-attr]
            raw_results.loc[("gryffindor", ["low", "very high"]), "value"].values
            == (group_sizes.loc[("gryffindor", [20, 80]), "value"] * step_number).values
        ).all()

    def _check_quidditch_wins(pop: pd.DataFrame, step_number: int) -> None:
        """We know that quidditch wins are stratified by 'familiar'
        and that each wizard with a banana slug familiar gains a point
        """
        assert set(pop["quidditch_wins"]) == set([0, 1])
        assert (pop.loc[pop["quidditch_wins"] != 0, "familiar"] == "banana_slug").all()
        group_sizes = pd.DataFrame(
            pop.groupby(["familiar"]).size().astype("float"), columns=[VALUE_COLUMN]
        )
        raw_results = sim._results._raw_results["quidditch_wins"]
        # We cannot use `equals` here because raw results have a MultiIndex where
        # each layer is a Category dtype but pop has object dtype for the relevant columns
        assert (
            raw_results[raw_results[VALUE_COLUMN] != 0].values
            == (group_sizes[group_sizes.index == "banana_slug"] * step_number).values
        ).all()

    components = [
        Hogwarts(),
        HousePointsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    pop = sim.get_population()
    _check_house_points(pop, step_number=1)
    _check_quidditch_wins(pop, step_number=1)

    sim.step()
    pop = sim.get_population()
    _check_house_points(pop, step_number=2)
    _check_quidditch_wins(pop, step_number=2)


def test_concatenating_observation_updates() -> None:
    """Test that concatenating observation raw results are being updated correctly."""
    components = [
        Hogwarts(),
        ExamScoreObserver(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    results_one_step = sim.get_results()["exam_score"]
    assert (results_one_step["exam_score"] == 10.0).all()
    sim.step()
    expected_two_steps = results_one_step.copy()
    expected_two_steps["exam_score"] = 20.0
    expected_two_steps["event_time"] = results_one_step["event_time"] + pd.Timedelta(
        days=sim.configuration.time.step_size
    )
    assert sim.get_results()["exam_score"].equals(
        pd.concat([results_one_step, expected_two_steps], axis=0).reset_index(drop=True)
    )


def test_update__raw_results_fully_filtered_pop() -> None:
    components = [
        Hogwarts(),
        FullyFilteredHousePointsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    # The FullyFilteredHousePointsObserver filters the population to a bogus
    # power level and so we should not be observing anything
    assert (sim._results._raw_results["house_points"][VALUE_COLUMN] == 0).all()
    sim.step()
    assert (sim._results._raw_results["house_points"][VALUE_COLUMN] == 0).all()


def test_update__raw_results_no_stratifications() -> None:
    components = [Hogwarts(), NoStratificationsQuidditchWinsObserver()]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    pop = sim.get_population()
    raw_results = sim._results._raw_results["no_stratifications_quidditch_wins"]
    assert raw_results.loc["all"][VALUE_COLUMN] == pop["quidditch_wins"].sum()
    sim.step()
    pop = sim.get_population()
    raw_results = sim._results._raw_results["no_stratifications_quidditch_wins"]
    assert raw_results.loc["all"][VALUE_COLUMN] == pop["quidditch_wins"].sum() * 2


def test_update__raw_results_extra_columns() -> None:
    """Test that raw results are updated correctly when the aggregator return
    contains multiple columns (i.e. not just a single 'value' column)
    """
    components = [Hogwarts(), HogwartsResultsStratifier(), MagicalAttributesObserver()]
    sim = InteractiveContext(configuration=HARRY_POTTER_CONFIG, components=components)
    sim.step()
    raw_results = sim._results._raw_results["magical_attributes"]
    assert (raw_results[["spell_power", "potion_power"]].values == [1, 1]).all()
    sim.step()
    raw_results = sim._results._raw_results["magical_attributes"]
    assert (raw_results[["spell_power", "potion_power"]].values == [2, 2]).all()



import pytest
from layered_config_tree.main import LayeredConfigTree
from pytest_mock import MockerFixture

from vivarium.framework.engine import Builder
from vivarium.framework.results.observer import Observer


class TestObserver(Observer):
    def register_observations(self, builder: Builder) -> None:
        pass


class TestDefaultObserverStratifications(Observer):
    def register_observations(self, builder: Builder) -> None:
        pass


class TestObserverStratifications(Observer):
    def register_observations(self, builder: Builder) -> None:
        pass

    @property
    def configuration_defaults(self) -> dict[str, str]:
        return {"foo": "bar"}


def test_observer_instantiation() -> None:
    observer = TestObserver()
    assert observer.name == "test_observer"


@pytest.mark.parametrize(
    "is_interactive, results_dir",
    [
        (False, "/some/results/dir"),
        (True, None),
    ],
)
def test_set_results_dir(
    is_interactive: bool, results_dir: str | None, mocker: MockerFixture
) -> None:
    builder = mocker.Mock()
    if is_interactive:
        builder.configuration = LayeredConfigTree()
    else:
        builder.configuration = LayeredConfigTree(
            {
                "output_data": {"results_directory": results_dir},
            }
        )

    observer = TestObserver()
    observer.set_results_dir(builder)

    assert observer.results_dir == results_dir



"""
============
Observations
============

An observation is a class object that records simulation results; they are responsible
for initializing, gathering, updating, and formatting results.

The provided :class:`Observation` class is an abstract base class that should
be subclassed by concrete observations. While there are no required abstract methods
to define when subclassing, the class does provide common attributes as well
as an `observe` method that determines whether to observe results for a given event.

At the highest level, an observation can be categorized as either an
:class:`UnstratifiedObservation` or a :class:`StratifiedObservation`. More specialized
implementations of these classes involve defining the various methods
provided as attributes to the parent class.

"""

from __future__ import annotations

import itertools
from abc import ABC
from collections.abc import Callable
from dataclasses import dataclass

import pandas as pd
from pandas.api.types import CategoricalDtype
from pandas.core.groupby.generic import DataFrameGroupBy

from vivarium.framework.event import Event
from vivarium.framework.results.stratification import Stratification

VALUE_COLUMN = "value"


@dataclass
class Observation(ABC):
    """An abstract base dataclass to be inherited by concrete observations.

    This class includes an :meth:`observe <observe>` method that determines whether
    to observe results for a given event.

    """

    name: str
    """Name of the observation. It will also be the name of the output results file
    for this particular observation."""
    pop_filter: str
    """A Pandas query filter string to filter the population down to the simulants
    who should be considered for the observation."""
    when: str
    """Name of the lifecycle phase the observation should happen. Valid values are:
    "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics"."""
    results_initializer: Callable[
        [tuple[str, ...] | None, list[Stratification]], pd.DataFrame
    ]
    """Method or function that initializes the raw observation results
    prior to starting the simulation. This could return, for example, an empty
    DataFrame or one with a complete set of stratifications as the index and
    all values set to 0.0."""
    results_gatherer: Callable[
        [
            pd.DataFrame | DataFrameGroupBy[tuple[str, ...] | str, bool],
            tuple[str, ...] | None,
        ],
        pd.DataFrame,
    ]
    """Method or function that gathers the new observation results."""
    results_updater: Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame]
    """Method or function that updates existing raw observation results with newly
    gathered results."""
    results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame]
    """Method or function that formats the raw observation results."""
    stratifications: tuple[str, ...] | None
    """Optional tuple of column names for the observation to stratify by."""
    to_observe: Callable[[Event], bool]
    """Method or function that determines whether to perform an observation on this Event."""

    def observe(
        self,
        event: Event,
        df: pd.DataFrame | DataFrameGroupBy[tuple[str, ...] | str, bool],
        stratifications: tuple[str, ...] | None,
    ) -> pd.DataFrame | None:
        """Determine whether to observe the given event, and if so, gather the results.

        Parameters
        ----------
        event
            The event to observe.
        df
            The population or population grouped by the stratifications.
        stratifications
            The stratifications to use for the observation.

        Returns
        -------
            The results of the observation.
        """
        if not self.to_observe(event):
            return None
        else:
            return self.results_gatherer(df, stratifications)


class UnstratifiedObservation(Observation):
    """Concrete class for observing results that are not stratified.

    The parent class `stratifications` are set to None and the `results_initializer`
    method is explicitly defined.

    Attributes
    ----------
    name
        Name of the observation. It will also be the name of the output results file
        for this particular observation.
    pop_filter
        A Pandas query filter string to filter the population down to the simulants who should
        be considered for the observation.
    when
        Name of the lifecycle phase the observation should happen. Valid values are:
        "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
    results_gatherer
        Method or function that gathers the new observation results.
    results_updater
        Method or function that updates existing raw observation results with newly gathered results.
    results_formatter
        Method or function that formats the raw observation results.
    to_observe
        Method or function that determines whether to perform an observation on this Event.

    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        results_gatherer: Callable[[pd.DataFrame], pd.DataFrame],
        results_updater: Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame],
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        def _wrap_results_gatherer(
            df: pd.DataFrame | DataFrameGroupBy[tuple[str, ...] | str, bool],
            _: tuple[str, ...] | None,
        ) -> pd.DataFrame:
            if isinstance(df, DataFrameGroupBy):
                raise TypeError(
                    "Must provide a dataframe to an UnstratifiedObservation. "
                    f"Provided DataFrameGroupBy instead."
                )
            return results_gatherer(df)

        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_initializer=self.create_empty_df,
            results_gatherer=_wrap_results_gatherer,
            results_updater=results_updater,
            results_formatter=results_formatter,
            stratifications=None,
            to_observe=to_observe,
        )

    @staticmethod
    def create_empty_df(
        requested_stratification_names: tuple[str, ...] | None,
        registered_stratifications: list[Stratification],
    ) -> pd.DataFrame:
        """Initialize an empty dataframe.

        Parameters
        ----------
        requested_stratification_names
            The names of the stratifications requested for this observation.
        registered_stratifications
            The list of all registered stratifications.

        Returns
        -------
            An empty DataFrame.
        """
        return pd.DataFrame()


class StratifiedObservation(Observation):
    """Concrete class for observing stratified results.

    The parent class `results_initializer` and `results_gatherer` methods are
    explicitly defined and stratification-specific attributes `aggregator_sources`
    and `aggregator` are added.

    Attributes
    ----------
    name
        Name of the observation. It will also be the name of the output results file
        for this particular observation.
    pop_filter
        A Pandas query filter string to filter the population down to the simulants who should
        be considered for the observation.
    when
        Name of the lifecycle phase the observation should happen. Valid values are:
        "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
    results_updater
        Method or function that updates existing raw observation results with newly gathered results.
    results_formatter
        Method or function that formats the raw observation results.
    stratifications
        Tuple of column names for the observation to stratify by. If empty,
        the observation is aggregated over the entire population.
    aggregator_sources
        List of population view columns to be used in the `aggregator`.
    aggregator
        Method or function that computes the quantity for this observation.
    to_observe
        Method or function that determines whether to perform an observation on this Event.

    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        results_updater: Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame],
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        stratifications: tuple[str, ...],
        aggregator_sources: list[str] | None,
        aggregator: Callable[[pd.DataFrame], float | pd.Series[float]],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_initializer=self.create_expanded_df,
            results_gatherer=self.get_complete_stratified_results,  # type: ignore [arg-type]
            results_updater=results_updater,
            results_formatter=results_formatter,
            stratifications=stratifications,
            to_observe=to_observe,
        )
        self.aggregator_sources = aggregator_sources
        self.aggregator = aggregator

    @staticmethod
    def create_expanded_df(
        requested_stratification_names: tuple[str, ...] | None,
        registered_stratifications: list[Stratification],
    ) -> pd.DataFrame:
        """Initialize a dataframe of 0s with complete set of stratifications as the index.

        Parameters
        ----------
        requested_stratification_names
            The names of the stratifications requested for this observation.
        registered_stratifications
            The list of all registered stratifications.

        Returns
        -------
            An empty DataFrame with the complete set of stratifications as the index.

        Notes
        -----
        If no stratifications are requested, then we are aggregating over the
        entire population and a single-row index named 'stratification' is created.
        """

        # Set up the complete index of all used stratifications
        if requested_stratification_names is not None:
            requested_and_registered_stratifications = [
                stratification
                for stratification in registered_stratifications
                if stratification.name in requested_stratification_names
            ]
        else:
            requested_and_registered_stratifications = []
        stratification_values = {
            stratification.name: stratification.categories
            for stratification in requested_and_registered_stratifications
        }
        if stratification_values:
            stratification_names = list(stratification_values.keys())
            df = pd.DataFrame(
                list(itertools.product(*stratification_values.values())),
                columns=stratification_names,
            ).astype(CategoricalDtype())
        else:
            # We are aggregating the entire population so create a single-row index
            stratification_names = ["stratification"]
            df = pd.DataFrame(["all"], columns=stratification_names).astype(
                CategoricalDtype()
            )

        # Initialize a zeros dataframe
        df[VALUE_COLUMN] = 0.0
        df = df.set_index(stratification_names)

        return df

    def get_complete_stratified_results(
        self,
        pop_groups: DataFrameGroupBy[str, bool],
        stratifications: tuple[str, ...],
    ) -> pd.DataFrame:
        """Gather results for this observation.

        Parameters
        ----------
        pop_groups
            The population grouped by the stratifications.
        stratifications
            The stratifications to use for the observation.

        Returns
        -------
            The results of the observation.
        """
        df = self._aggregate(pop_groups, self.aggregator_sources, self.aggregator)
        df = self._format(df)
        df = self._expand_index(df)
        if not list(stratifications):
            df.index.name = "stratification"
        return df

    @staticmethod
    def _aggregate(
        pop_groups: DataFrameGroupBy[str, bool],
        aggregator_sources: list[str] | None,
        aggregator: Callable[[pd.DataFrame], float | pd.Series[float]],
    ) -> pd.Series[float] | pd.DataFrame:
        """Apply the `aggregator` to the population groups and their
        `aggregator_sources` columns.
        """
        aggregates = (
            pop_groups[aggregator_sources].apply(aggregator).fillna(0.0)  # type: ignore [arg-type]
            if aggregator_sources
            else pop_groups.apply(aggregator)  # type: ignore [arg-type]
        ).astype(float)
        return aggregates

    @staticmethod
    def _format(aggregates: pd.Series[float] | pd.DataFrame) -> pd.DataFrame:
        """Convert the results to a pandas DataFrame if necessary and ensure the
        results column name is 'value'.
        """
        df = pd.DataFrame(aggregates) if isinstance(aggregates, pd.Series) else aggregates
        if df.shape[1] == 1:
            df.rename(columns={df.columns[0]: "value"}, inplace=True)
        return df

    @staticmethod
    def _expand_index(aggregates: pd.DataFrame) -> pd.DataFrame:
        """Include all stratifications in the results by filling missing values with 0."""
        full_idx = (
            pd.MultiIndex.from_product(aggregates.index.levels)
            if isinstance(aggregates.index, pd.MultiIndex)
            else aggregates.index
        )
        aggregates = aggregates.reindex(full_idx).fillna(0.0)
        return aggregates


class AddingObservation(StratifiedObservation):
    """Concrete class for observing additive and stratified results.

    The parent class `results_updater` method is explicitly defined and
    stratification-specific attributes `aggregator_sources` and `aggregator` are added.

    Attributes
    ----------
    name
        Name of the observation. It will also be the name of the output results file
        for this particular observation.
    pop_filter
        A Pandas query filter string to filter the population down to the simulants who should
        be considered for the observation.
    when
        Name of the lifecycle phase the observation should happen. Valid values are:
        "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
    results_formatter
        Method or function that formats the raw observation results.
    stratifications
        Optional tuple of column names for the observation to stratify by.
    aggregator_sources
        List of population view columns to be used in the `aggregator`.
    aggregator
        Method or function that computes the quantity for this observation.
    to_observe
        Method or function that determines whether to perform an observation on this Event.

    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        stratifications: tuple[str, ...],
        aggregator_sources: list[str] | None,
        aggregator: Callable[[pd.DataFrame], float | pd.Series[float]],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_updater=self.add_results,
            results_formatter=results_formatter,
            stratifications=stratifications,
            aggregator_sources=aggregator_sources,
            aggregator=aggregator,
            to_observe=to_observe,
        )

    @staticmethod
    def add_results(
        existing_results: pd.DataFrame, new_observations: pd.DataFrame
    ) -> pd.DataFrame:
        """Add newly-observed results to the existing results.

        Parameters
        ----------
        existing_results
            The existing results DataFrame.
        new_observations
            The new observations DataFrame.

        Returns
        -------
            The new results added to the existing results.

        Notes
        -----
        If the new observations contain columns not present in the existing results,
        the columns are added to the DataFrame and initialized with 0.0s.
        """
        updated_results = existing_results.copy()
        # Look for extra columns in the new_observations and initialize with 0.
        extra_cols = [
            c for c in new_observations.columns if c not in existing_results.columns
        ]
        if extra_cols:
            updated_results[extra_cols] = 0.0
        for col in new_observations.columns:
            updated_results[col] += new_observations[col]
        return updated_results


class ConcatenatingObservation(UnstratifiedObservation):
    """Concrete class for observing concatenating (and by extension, unstratified) results.

    The parent class `results_gatherer` and `results_updater` methods are explicitly
    defined and attribute `included_columns` is added.

    Attributes
    ----------
    name
        Name of the observation. It will also be the name of the output results file
        for this particular observation.
    pop_filter
        A Pandas query filter string to filter the population down to the simulants who should
        be considered for the observation.
    when
        Name of the lifecycle phase the observation should happen. Valid values are:
        "time_step__prepare", "time_step", "time_step__cleanup", or "collect_metrics".
    included_columns
        Columns to include in the observation
    results_formatter
        Method or function that formats the raw observation results.
    to_observe
        Method or function that determines whether to perform an observation on this Event.

    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        included_columns: list[str],
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_gatherer=self.get_results_of_interest,
            results_updater=self.concatenate_results,
            results_formatter=results_formatter,
            to_observe=to_observe,
        )
        self.included_columns = included_columns

    def get_results_of_interest(self, pop: pd.DataFrame) -> pd.DataFrame:
        """Return the population with only the `included_columns`."""
        return pop[self.included_columns]

    @staticmethod
    def concatenate_results(
        existing_results: pd.DataFrame, new_observations: pd.DataFrame
    ) -> pd.DataFrame:
        """Concatenate the existing results with the new observations.

        Parameters
        ----------
        existing_results
            The existing results.
        new_observations
            The new observations.

        Returns
        -------
            The new results concatenated to the existing results.
        """
        if existing_results.empty:
            return new_observations
        return pd.concat([existing_results, new_observations], axis=0).reset_index(drop=True)



from __future__ import annotations

import warnings
from collections.abc import Callable, Iterable, Sequence
from typing import TYPE_CHECKING, Any, TypeVar

from vivarium.framework.event import Event
from vivarium.framework.resource import Resource
from vivarium.framework.values.combiners import ValueCombiner, replace_combiner
from vivarium.framework.values.pipeline import Pipeline
from vivarium.framework.values.post_processors import PostProcessor, rescale_post_processor
from vivarium.manager import Interface, Manager

if TYPE_CHECKING:
    from vivarium import Component
    from vivarium.framework.engine import Builder

T = TypeVar("T")


class ValuesManager(Manager):
    """Manager for the dynamic value system."""

    def __init__(self) -> None:
        # Pipelines are lazily initialized by _register_value_producer
        self._pipelines: dict[str, Pipeline] = {}

    @property
    def name(self) -> str:
        return "values_manager"

    def setup(self, builder: Builder) -> None:
        self.logger = builder.logging.get_logger(self.name)
        self.step_size = builder.time.step_size()
        self.simulant_step_sizes = builder.time.simulant_step_sizes()
        builder.event.register_listener("post_setup", self.on_post_setup)

        self.resources = builder.resources
        self.add_constraint = builder.lifecycle.add_constraint

        builder.lifecycle.add_constraint(self.register_value_producer, allow_during=["setup"])
        builder.lifecycle.add_constraint(self.register_value_modifier, allow_during=["setup"])

    def on_post_setup(self, _event: Event) -> None:
        """Finalizes dependency structure for the pipelines."""
        # Unsourced pipelines might occur when generic components register
        # modifiers to values that aren't required in a simulation.
        unsourced_pipelines = [p for p, v in self._pipelines.items() if not v.source]
        if unsourced_pipelines:
            self.logger.warning(f"Unsourced pipelines: {unsourced_pipelines}")

        # register_value_producer and register_value_modifier record the
        # dependency structure for the pipeline source and pipeline modifiers,
        # respectively. We don't have enough information to record the
        # dependency structure for the pipeline itself until now, where
        # we say the pipeline value depends on its source and all its
        # modifiers.
        for name, pipe in self._pipelines.items():
            self.resources.add_resources(
                pipe.component, [pipe], [pipe.source] + list(pipe.mutators)
            )

    def register_value_producer(
        self,
        value_name: str,
        source: Callable[..., Any],
        # TODO [MIC-5452]: all calls should have a component
        component: Component | None = None,
        requires_columns: Iterable[str] = (),
        requires_values: Iterable[str] = (),
        requires_streams: Iterable[str] = (),
        required_resources: Sequence[str | Resource] = (),
        preferred_combiner: ValueCombiner = replace_combiner,
        preferred_post_processor: PostProcessor | None = None,
    ) -> Pipeline:
        """Marks a ``Callable`` as the producer of a named value.

        See Also
        --------
            :meth:`ValuesInterface.register_value_producer`
        """
        self.logger.debug(f"Registering value pipeline {value_name}")
        pipeline = self.get_value(value_name)
        pipeline.set_attributes(
            component,
            source,
            preferred_combiner,
            preferred_post_processor,
            self,
        )

        # The resource we add here is just the pipeline source.
        # The value will depend on the source and its modifiers, and we'll
        # declare that resource at post-setup once all sources and modifiers
        # are registered.
        dependencies = self._convert_dependencies(
            source, requires_columns, requires_values, requires_streams, required_resources
        )
        self.resources.add_resources(pipeline.component, [pipeline.source], dependencies)

        self.add_constraint(
            pipeline._call, restrict_during=["initialization", "setup", "post_setup"]
        )

        return pipeline

    def register_value_modifier(
        self,
        value_name: str,
        modifier: Callable[..., Any],
        # TODO [MIC-5452]: all calls should have a component
        component: Component | Manager | None = None,
        requires_columns: Iterable[str] = (),
        requires_values: Iterable[str] = (),
        requires_streams: Iterable[str] = (),
        required_resources: Sequence[str | Resource] = (),
    ) -> None:
        """Marks a ``Callable`` as the modifier of a named value.

        Parameters
        ----------
        value_name :
            The name of the dynamic value pipeline to be modified.
        modifier :
            A function that modifies the source of the dynamic value pipeline
            when called. If the pipeline has a ``replace_combiner``, the
            modifier should accept the same arguments as the pipeline source
            with an additional last positional argument for the results of the
            previous stage in the pipeline. For the ``list_combiner`` strategy,
            the pipeline modifiers should have the same signature as the pipeline
            source.
        component
            The component that is registering the value modifier.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the pipeline modifier
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the pipeline modifier is called.
        requires_streams
            A list of the randomness streams that need to be properly sourced
            before the pipeline modifier is called.
        required_resources
            A list of resources that need to be properly sourced before the
            pipeline modifier is called.  This is a list of strings, pipeline
            names, or randomness streams.
        """
        pipeline = self.get_value(value_name)
        value_modifier = pipeline.get_value_modifier(modifier, component)
        self.logger.debug(f"Registering {value_modifier.name} as modifier to {value_name}")

        dependencies = self._convert_dependencies(
            modifier, requires_columns, requires_values, requires_streams, required_resources
        )
        self.resources.add_resources(component, [value_modifier], dependencies)

    def get_value(self, name: str) -> Pipeline:
        """Retrieve the pipeline representing the named value.

        Parameters
        ----------
        name
            Name of the pipeline to return.

        Returns
        -------
            A callable reference to the named pipeline.  The pipeline arguments
            should be identical to the arguments to the pipeline source
            (frequently just a :class:`pandas.Index` representing the
            simulants).
        """
        pipeline = self._pipelines.get(name) or Pipeline(name)
        self._pipelines[name] = pipeline
        return pipeline

    @staticmethod
    def _convert_dependencies(
        func: Callable[..., Any],
        requires_columns: Iterable[str],
        requires_values: Iterable[str],
        requires_streams: Iterable[str],
        required_resources: Iterable[str | Resource],
    ) -> Iterable[str | Resource]:
        if isinstance(func, Pipeline):
            # The dependencies of the pipeline itself will have been declared
            # when the pipeline was registered.
            return [func]

        if requires_columns or requires_values or requires_streams:
            warnings.warn(
                "Specifying requirements individually is deprecated. You should "
                "specify them using the 'required_resources' argument instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            if required_resources:
                raise ValueError(
                    "If requires_columns, requires_values, or requires_streams"
                    " are provided, requirements must be empty."
                )

            return (
                list(requires_columns)
                + [Resource("value", name, None) for name in requires_values]
                + [Resource("stream", name, None) for name in requires_streams]
            )
        else:
            return required_resources

    def keys(self) -> Iterable[str]:
        """Get an iterable of pipeline names."""
        return self._pipelines.keys()

    def items(self) -> Iterable[tuple[str, Pipeline]]:
        """Get an iterable of name, pipeline tuples."""
        return self._pipelines.items()

    def values(self) -> Iterable[Pipeline]:
        """Get an iterable of all pipelines."""
        return self._pipelines.values()

    def __contains__(self, item: str) -> bool:
        return item in self._pipelines

    def __iter__(self) -> Iterable[str]:
        return iter(self._pipelines)

    def __repr__(self) -> str:
        return "ValuesManager()"


class ValuesInterface(Interface):
    """Public interface for the simulation values management system.

    The values system provides tools to build up a value across many
    components, allowing users to build components that focus on small groups
    of simulant attributes.

    """

    def __init__(self, manager: ValuesManager) -> None:
        self._manager = manager

    def register_value_producer(
        self,
        value_name: str,
        source: Callable[..., Any],
        # TODO [MIC-5452]: all calls should have a component
        component: Component | None = None,
        requires_columns: Iterable[str] = (),
        requires_values: Iterable[str] = (),
        requires_streams: Iterable[str] = (),
        required_resources: Sequence[str | Resource] = (),
        preferred_combiner: ValueCombiner = replace_combiner,
        preferred_post_processor: PostProcessor | None = None,
    ) -> Pipeline:
        """Marks a ``Callable`` as the producer of a named value.

        Parameters
        ----------
        value_name
            The name of the new dynamic value pipeline.
        source
            A callable source for the dynamic value pipeline.
        component
            The component that is registering the value producer.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the pipeline source
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the pipeline source is called.
        requires_streams
            A list of the randomness streams that need to be properly sourced
            before the pipeline source is called.
        required_resources
            A list of resources that need to be properly sourced before the
            pipeline source is called.  This is a list of strings, pipeline
            names, or randomness streams.
        preferred_combiner
            A strategy for combining the source and the results of any calls
            to mutators in the pipeline. ``vivarium`` provides the strategies
            ``replace_combiner`` (the default) and ``list_combiner``, which
            are importable from ``vivarium.framework.values``.  Client code
            may define additional strategies as necessary.
        preferred_post_processor
            A strategy for processing the final output of the pipeline.
            ``vivarium`` provides the strategies ``rescale_post_processor``
            and ``union_post_processor`` which are importable from
            ``vivarium.framework.values``.  Client code may define additional
            strategies as necessary.

        Returns
        -------
            A callable reference to the named dynamic value pipeline.
        """
        return self._manager.register_value_producer(
            value_name,
            source,
            component,
            requires_columns,
            requires_values,
            requires_streams,
            required_resources,
            preferred_combiner,
            preferred_post_processor,
        )

    def register_rate_producer(
        self,
        rate_name: str,
        source: Callable[..., Any],
        # TODO [MIC-5452]: all calls should have a component
        component: Component | None = None,
        requires_columns: Iterable[str] = (),
        requires_values: Iterable[str] = (),
        requires_streams: Iterable[str] = (),
        required_resources: Sequence[str | Resource] = (),
    ) -> Pipeline:
        """Marks a ``Callable`` as the producer of a named rate.

        This is a convenience wrapper around ``register_value_producer`` that
        makes sure rate data is appropriately scaled to the size of the
        simulation time step.  It is equivalent to
        ``register_value_producer(value_name, source,
        preferred_combiner=replace_combiner,
        preferred_post_processor=rescale_post_processor)``

        Parameters
        ----------
        rate_name
            The name of the new dynamic rate pipeline.
        source
            A callable source for the dynamic rate pipeline.
        component
            The component that is registering the rate producer.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the pipeline source
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the pipeline source is called.
        requires_streams
            A list of the randomness streams that need to be properly sourced
            before the pipeline source is called.
        required_resources
            A list of resources that need to be properly sourced before the
            pipeline source is called.  This is a list of strings, pipeline
            names, or randomness streams.

        Returns
        -------
            A callable reference to the named dynamic rate pipeline.
        """
        return self.register_value_producer(
            rate_name,
            source,
            component,
            requires_columns,
            requires_values,
            requires_streams,
            required_resources,
            preferred_post_processor=rescale_post_processor,
        )

    def register_value_modifier(
        self,
        value_name: str,
        modifier: Callable[..., Any],
        # TODO [MIC-5452]: all calls should have a component
        component: Component | Manager | None = None,
        requires_columns: Iterable[str] = (),
        requires_values: Iterable[str] = (),
        requires_streams: Iterable[str] = (),
        required_resources: Sequence[str | Resource] = (),
    ) -> None:
        """Marks a ``Callable`` as the modifier of a named value.

        Parameters
        ----------
        value_name :
            The name of the dynamic value pipeline to be modified.
        modifier :
            A function that modifies the source of the dynamic value pipeline
            when called. If the pipeline has a ``replace_combiner``, the
            modifier should accept the same arguments as the pipeline source
            with an additional last positional argument for the results of the
            previous stage in the pipeline. For the ``list_combiner`` strategy,
            the pipeline modifiers should have the same signature as the pipeline
            source.
        component
            The component that is registering the value modifier.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the pipeline modifier
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the pipeline modifier is called.
        requires_streams
            A list of the randomness streams that need to be properly sourced
            before the pipeline modifier is called.
        required_resources
            A list of resources that need to be properly sourced before the
            pipeline modifier is called.  This is a list of strings, pipeline
            names, or randomness streams.
        """
        self._manager.register_value_modifier(
            value_name,
            modifier,
            component,
            requires_columns,
            requires_values,
            requires_streams,
            required_resources,
        )

    def get_value(self, name: str) -> Pipeline:
        """Retrieve the pipeline representing the named value.

        Parameters
        ----------
        name
            Name of the pipeline to return.

        Returns
        -------
            A callable reference to the named pipeline.  The pipeline arguments
            should be identical to the arguments to the pipeline source
            (frequently just a :class:`pandas.Index` representing the
            simulants).

        """
        return self._manager.get_value(name)



from __future__ import annotations

import math
from typing import Any, Generator
from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pytest
import pytest_mock
from layered_config_tree import LayeredConfigTree

from tests.helpers import Listener, MockComponentA, MockComponentB, MockGenericComponent
from vivarium.component import Component
from vivarium.framework.engine import Builder, SimulationContext
from vivarium.framework.event import Event
from vivarium.framework.results.observer import Observer
from vivarium.framework.time import SimulationClock, get_time_stamp
from vivarium.framework.utilities import from_yearly
from vivarium.framework.values import ValuesManager, rescale_post_processor
from vivarium.types import ClockStepSize, ClockTime, NumberLike


@pytest.fixture
def builder(mocker: pytest_mock.MockFixture) -> Any:
    builder = mocker.MagicMock()
    manager = ValuesManager()
    manager.setup(builder)
    builder.value.register_value_producer = manager.register_value_producer
    builder.value.register_value_modifier = manager.register_value_modifier
    return builder


@pytest.fixture
def components() -> list[Component | Listener]:
    return [
        MockComponentA("gretchen", "whimsy"),
        Listener("listener"),
        MockComponentB("spoon", "antelope", "23"),
    ]


def validate_step_column_is_pipeline(sim: SimulationContext) -> None:
    """Ensure that the pipeline and column step sizes are aligned"""
    step_pipeline = sim._values.get_value("simulant_step_size")(sim.get_population().index)
    assert sim._population._population is not None
    step_column = sim._population._population.step_size
    assert np.all(step_pipeline == step_column)


def validate_index_aligned(
    sim: SimulationContext, expected_active_simulants: pd.Index[int]
) -> None:
    """Ensure that the active simulants are as expected BEFORE a step"""
    active_simulants = sim._clock.get_active_simulants(
        get_full_pop_index(sim), sim._clock.event_time
    )

    assert active_simulants.equals(expected_active_simulants)
    assert active_simulants.difference(expected_active_simulants).empty


def validate_event_indexes(listener: Listener, expected_simulants: pd.Index[int]) -> None:
    """Make sure AFTER a step, that simulants were included in the right events"""
    for index in listener.event_indexes.values():
        assert index is not None
        assert index.equals(expected_simulants)


def take_step(sim: SimulationContext) -> ClockStepSize:
    old_time = sim._clock.time
    sim.step()
    new_time = sim._clock.time
    return new_time - old_time  # type: ignore [operator]


def get_full_pop_index(sim: SimulationContext) -> pd.Index[int]:
    return sim.get_population().index


def get_index_by_parity(index: pd.Index[int], parity: str) -> pd.Index[int]:
    if parity == "evens":
        return index[index % 2 == 0]
    elif parity == "odds":
        return index[index % 2 == 1]
    else:
        return index


def get_pop_by_parity(sim: SimulationContext, parity: str) -> pd.DataFrame:
    pop = sim.get_population()
    return pop.loc[get_index_by_parity(pop.index, parity)]


def pipeline_by_parity(
    sim: SimulationContext, step_modifiers: dict[str, int], parity: str
) -> pd.Series[Any]:
    if parity == "all":
        modified_series: pd.Series[float] = pd.concat(
            [
                pipeline_by_parity(sim, step_modifiers, "evens"),
                pipeline_by_parity(sim, step_modifiers, "odds"),
            ]
        ).sort_index()
        return modified_series
    return pd.Series(
        from_yearly(1.75, pd.Timedelta(days=step_modifiers[parity])),
        index=get_index_by_parity(get_full_pop_index(sim), parity),
    )


def take_step_and_validate(
    sim: SimulationContext,
    listener: Listener,
    expected_simulants: pd.Index[int],
    expected_step_size_days: int,
) -> None:
    """Take a step, and ensure that we included the right simulants, with the right step size"""
    ## Check Before Timestep
    validate_index_aligned(sim, expected_simulants)
    ## Check Timestep
    assert take_step(sim) == pd.Timedelta(days=expected_step_size_days)
    ## Check After
    validate_event_indexes(listener, expected_simulants)


class StepModifier(MockGenericComponent):
    """This mock component modifies the step size of simulants based on their index
    Odd simulants get one step size, and even simulants get another.
    """

    def __init__(
        self,
        name: str,
        step_modifier_even: int | float,
        step_modifier_odd: int | float | None = None,
        modified_simulants: str = "all",
    ):
        super().__init__(name)
        self.step_modifier_even = step_modifier_even
        self.step_modifier_odd = (
            step_modifier_odd if step_modifier_odd else step_modifier_even
        )
        self.modified_simulants = modified_simulants

    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        builder.time.register_step_size_modifier(self.modify_step)

    def modify_step(self, index: pd.Index[int]) -> pd.Series[ClockStepSize]:
        step_sizes = pd.Series(pd.Timedelta(days=1), index=index)
        step_sizes.loc[get_index_by_parity(index, "evens")] = pd.Timedelta(
            days=self.step_modifier_even
        )
        step_sizes.loc[get_index_by_parity(index, "odds")] = pd.Timedelta(
            days=self.step_modifier_odd
        )
        modified_steps = step_sizes.loc[get_index_by_parity(index, self.modified_simulants)]
        return modified_steps  # type: ignore [return-value]


class StepModifierWithRatePipeline(StepModifier):
    """
    Add a test pipeline that is registered, whose value is cached to
    self.ts_pipeline_value every timestep. This is meant to ensure that the
    value of a pipeline on a previous timestep was appropriately rescaled
    to the step that was actually taken.
    """

    def __init__(
        self,
        name: str,
        step_modifier_even: float | int,
        step_modifier_odd: float | int | None = None,
        modified_simulants: str = "all",
    ) -> None:
        super().__init__(name, step_modifier_even, step_modifier_odd, modified_simulants)
        self.ts_pipeline_value = None

    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        self.rate_pipeline = builder.value.register_value_producer(
            f"test_rate_{self.name}",
            source=lambda idx: pd.Series(1.75, index=idx),
            preferred_post_processor=rescale_post_processor,
        )

    def on_time_step(self, event: Event) -> None:
        self.ts_pipeline_value = self.rate_pipeline(event.index)


class StepModifierWithUntracking(StepModifierWithRatePipeline):
    """Add an event step that untracks/tracks even simulants every timestep"""

    @property
    def columns_required(self) -> list[str]:
        return ["tracked"]

    def on_time_step(self, event: Event) -> None:
        super().on_time_step(event)
        evens = self.population_view.get(event.index).loc[
            get_index_by_parity(event.index, "evens")
        ]
        evens["tracked"] = False
        self.population_view.update(evens)


class StepModifierWithMovement(StepModifierWithRatePipeline):
    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        self.move_simulants_to_end = builder.time.move_simulants_to_end()

    def on_time_step(self, event: Event) -> None:
        super().on_time_step(event)
        self.move_simulants_to_end(get_index_by_parity(event.index, "evens"))


@pytest.mark.parametrize("varied_step_size", [True, False])
def test_basic_iteration(
    base_config: LayeredConfigTree,
    components: list[Component | Listener],
    varied_step_size: bool,
) -> None:
    """Ensure that the basic iteration of the simulation works as expected.
    The step size should always be 1 in this case, the whole population should
    be updated, and the pipeline step value should always match the column step value.
    """
    if varied_step_size:
        components.append(StepModifier("step_modifier", 1))
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if hasattr(c, "args") and "listener" in c.args][0]
    assert isinstance(listener, Listener)
    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)
    assert sim._clock.time == get_time_stamp(sim.configuration.time.start)
    assert sim._clock.step_size == pd.Timedelta(days=1)
    ## Ensure that we don't have a pop view (and by extension, don't vary clocks)
    ## If no components modify the step size.
    assert bool(sim._clock._individual_clocks) == varied_step_size

    for _ in range(2):
        # After initialization, all simulants should be aligned to event times
        # After a step (and no step adjustments), simulants should still be aligned
        if varied_step_size:
            validate_step_column_is_pipeline(sim)
            assert np.all(
                sim._clock.simulant_next_event_times(full_pop_index) == sim._clock.event_time
            )
            assert np.all(
                sim._clock.simulant_step_sizes(full_pop_index) == sim._clock.step_size
            )
        take_step_and_validate(
            sim, listener, expected_simulants=full_pop_index, expected_step_size_days=1
        )


def test_empty_active_pop(
    base_config: LayeredConfigTree, components: list[Component | Listener]
) -> None:
    """Make sure that if we have no active simulants, we still take a step, given
    by the minimum step size."""
    components.append(StepModifier("step_modifier", 1))
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if hasattr(c, "args") and "listener" in c.args][0]
    assert isinstance(listener, Listener)
    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)

    ## Force a next event time update without updating step sizes.
    ## This ensures (against the current implementation) that we will have a timestep
    ## that has no simulants aligned. Check that we do the minimum timestep update.
    assert sim._population._population is not None
    sim._population._population.next_event_time += pd.Timedelta(days=1)
    ## First Step
    validate_step_column_is_pipeline(sim)
    take_step_and_validate(
        sim, listener, expected_simulants=pd.Index([]), expected_step_size_days=1
    )

    ## Second Timestep
    validate_step_column_is_pipeline(sim)
    take_step_and_validate(
        sim, listener, expected_simulants=full_pop_index, expected_step_size_days=1
    )


@pytest.mark.parametrize(
    "step_modifier_even,step_modifier_odd", [(0.5, 0.5), (1, 1), (2, 2), (4.5, 4)]
)
def test_skip_iterations(
    base_config: LayeredConfigTree,
    step_modifier_even: float | int,
    step_modifier_odd: float | int,
) -> None:
    """Test that if everyone has some (non-minimum) step size, the global step adjusts to match"""
    listener = Listener("listener")
    sim = SimulationContext(
        base_config,
        [StepModifier("step_modifier", step_modifier_even, step_modifier_odd), listener],
    )
    expected_step_size = math.floor(max([step_modifier_even, 1]))
    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)

    ## Go through a couple simulant step cycles
    for _ in range(2):
        ## Everyone should update, but the step size should change
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=full_pop_index,
            expected_step_size_days=expected_step_size,
        )


def test_uneven_steps(base_config: LayeredConfigTree) -> None:
    """Test that if we have a mix of step sizes, we take steps in accordance
    to reach all simulants' next event times in the fewest steps.
    """

    listener = Listener("listener")
    step_modifiers = {"evens": 3, "odds": 7}
    step_modifier_component = StepModifierWithRatePipeline(
        "step_modifier", step_modifiers["evens"], step_modifiers["odds"]
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## With step sizes of 3 and 7, we need 3 steps, then 3 more, then one to get 7, then two more to get
    ## 9, etc.
    correct_step_sizes = [3, 3, 1, 2, 3, 2, 1, 3, 3]
    groups = ["evens", "evens", "odds", "evens", "evens", "odds", "evens", "evens", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 21
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )

        sample_pipeline = step_modifier_component.ts_pipeline_value
        assert sample_pipeline is not None
        assert sample_pipeline.index.equals(get_pop_by_parity(sim, group).index)
        assert np.all(sample_pipeline == pipeline_by_parity(sim, step_modifiers, group))


def test_partial_modification(base_config: LayeredConfigTree) -> None:
    """Test that if we have one modifier that doesn't apply to all simulants,
    we choose the standard value for unmodified simulants.
    """

    listener = Listener("listener")
    ## Define odds for validation, but don't pass it into the step modifier
    step_modifiers = {"evens": 3, "odds": 1}
    step_modifier_component = StepModifierWithRatePipeline(
        "step_modifier", step_modifiers["evens"], modified_simulants="evens"
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## We should update odds with default each time, and update evens every third step.
    correct_step_sizes = [1, 1, 1, 1, 1, 1, 1, 1, 1]
    groups = ["odds", "odds", "all", "odds", "odds", "all", "odds", "odds", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 21
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )

        sample_pipeline = step_modifier_component.ts_pipeline_value
        assert sample_pipeline is not None
        assert sample_pipeline.index.equals(get_pop_by_parity(sim, group).index)
        assert np.all(sample_pipeline == pipeline_by_parity(sim, step_modifiers, group))


def test_standard_step_size(base_config: LayeredConfigTree) -> None:
    """Test that if we have one modifier that doesn't apply to all simulants,
    we choose the standard value for unmodified simulants.
    """

    base_config.update({"configuration": {"time": {"standard_step_size": 5}}})
    listener = Listener("listener")
    ## Define odds for validation, but don't pass it into the step modifier
    step_modifiers = {"evens": 3, "odds": 5}
    step_modifier_component = StepModifierWithRatePipeline(
        "step_modifier", step_modifiers["evens"], modified_simulants="evens"
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## We should give odds standard size, and evens 3 days.
    correct_step_sizes = [3, 2, 1, 3, 1, 2, 3]
    groups = ["evens", "odds", "evens", "evens", "odds", "evens", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 15
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )

        sample_pipeline = step_modifier_component.ts_pipeline_value
        assert sample_pipeline is not None
        assert sample_pipeline.index.equals(get_pop_by_parity(sim, group).index)
        assert np.all(sample_pipeline == pipeline_by_parity(sim, step_modifiers, group))


def test_multiple_modifiers(base_config: LayeredConfigTree) -> None:
    """Test that if we have a mix of step sizes, we take steps in accordance
    to reach all simulants' next event times in the fewest steps.
    """

    listener = Listener("listener")
    step_modifiers = {"evens": 3, "odds": 7}
    step_modifier_A = StepModifier(
        "step_modifier_A", step_modifiers["evens"], modified_simulants="evens"
    )
    step_modifier_B = StepModifier(
        "step_modifier_B",
        step_modifiers["evens"],
        step_modifiers["odds"],
        modified_simulants="odds",
    )
    sim = SimulationContext(
        base_config,
        [step_modifier_A, step_modifier_B, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    ## With step sizes of 3 and 7, we need 3 steps, then 3 more, then one to get 7, then two more to get
    ## 9, etc.
    correct_step_sizes = [3, 3, 1, 2, 3, 2, 1, 3, 3]
    groups = ["evens", "evens", "odds", "evens", "evens", "odds", "evens", "evens", "all"]

    ## Ensure that steps and active simulants are correct through one cycle of 21
    for correct_step_size, group in zip(correct_step_sizes, groups):
        validate_step_column_is_pipeline(sim)
        take_step_and_validate(
            sim,
            listener,
            expected_simulants=get_pop_by_parity(sim, group).index,
            expected_step_size_days=correct_step_size,
        )


def test_untracked_simulants(base_config: LayeredConfigTree) -> None:
    """Test that untracked simulants are always included in event indices, and are
    basically treated the same as any other simulant."""
    base_config.update({"configuration": {"time": {"standard_step_size": 7}}})
    listener = Listener("listener")
    step_modifier_component = StepModifierWithUntracking("step_modifier", 3)
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)

    for _ in range(2):
        take_step_and_validate(sim, listener, full_pop_index, expected_step_size_days=3)
        assert step_modifier_component.ts_pipeline_value is not None
        assert step_modifier_component.ts_pipeline_value.index.equals(full_pop_index)


def test_move_simulants_to_end(base_config: LayeredConfigTree) -> None:
    """Ensure that we move simulants' next event time to the end of the simulation, if they are even."""
    base_config.update({"configuration": {"time": {"standard_step_size": 7}}})
    listener = Listener("listener")
    step_modifier_component = StepModifierWithMovement("step_modifier", 3)
    sim = SimulationContext(
        base_config,
        [step_modifier_component, listener],
    )

    sim.setup()
    sim.initialize_simulants()
    full_pop_index = get_full_pop_index(sim)
    odds = get_index_by_parity(full_pop_index, "odds")
    evens = get_index_by_parity(full_pop_index, "evens")
    take_step_and_validate(sim, listener, full_pop_index, expected_step_size_days=3)
    assert step_modifier_component.ts_pipeline_value is not None
    assert step_modifier_component.ts_pipeline_value.index.equals(full_pop_index)
    assert np.all(
        sim._clock.simulant_next_event_times(evens)
        == sim._clock.stop_time + sim._clock.minimum_step_size
    )

    for _ in range(2):
        take_step_and_validate(sim, listener, odds, expected_step_size_days=3)
        assert step_modifier_component.ts_pipeline_value is not None
        assert step_modifier_component.ts_pipeline_value.index.equals(odds)


def test_step_size_post_processor(builder: MagicMock) -> None:
    """Test that step size post-processor chooses the minimum modified step, or minimum global step,
    whichever is larger."""
    index = pd.Index(range(10))
    clock = SimulationClock()
    clock._standard_step_size = clock._minimum_step_size = pd.Timedelta(days=2)
    clock.setup(builder)

    ## Add modifier that set the step size to 7 for even indices and 5 for odd indices
    clock.register_step_modifier(
        lambda idx: pd.Series(
            [pd.Timedelta(days=7) if i % 2 == 0 else pd.Timedelta(days=5) for i in idx],
            index=idx,
        )
    )
    ## Add modifier that sets the step size to 9 for all simulants
    clock.register_step_modifier(lambda idx: pd.Series(pd.Timedelta(days=9), index=idx))
    value = clock._step_size_pipeline(index)
    evens = value.iloc[lambda x: x.index % 2 == 0]
    odds = value.iloc[lambda x: x.index % 2 == 1]

    ## The second modifier shouldn't have an effect
    assert np.all(evens == pd.Timedelta(days=6))
    assert np.all(odds == pd.Timedelta(days=4))

    clock.register_step_modifier(lambda idx: pd.Series(pd.Timedelta(days=0.5), index=idx))
    value = clock._step_size_pipeline(index)
    assert np.all(value == pd.Timedelta(days=2))


@pytest.mark.parametrize("end_day", [31, 23])
def test_time_steps_remaining(base_config: LayeredConfigTree, end_day: int) -> None:
    UselessComponent = MockGenericComponent("Placeholder")
    base_config.update(
        {
            "configuration": {
                "time": {
                    "start": {"year": 2021, "month": 1, "day": 1},
                    "end": {"year": 2021, "month": 1, "day": end_day},
                    "step_size": 10,  # Days
                },
            },
        },
    )
    sim = SimulationContext(
        base_config,
        [UselessComponent],
    )
    sim.setup()
    sim.initialize_simulants()
    assert sim.get_number_of_steps_remaining() == 3
    sim.step()
    assert sim.get_number_of_steps_remaining() == 2


@pytest.mark.parametrize("end", [31, 23])
def test_simple_clock_time_steps_remaining(base_config: LayeredConfigTree, end: int) -> None:
    UselessComponent = MockGenericComponent("Placeholder")
    base_config.update(
        {
            "configuration": {
                "time": {
                    "start": 1,
                    "end": end,
                    "step_size": 10,
                },
            },
            "plugins": {
                "required": {
                    "clock": {
                        "controller": "vivarium.framework.time.SimpleClock",
                    },
                }
            },
        },
    )
    sim = SimulationContext(
        base_config,
        [UselessComponent],
    )
    sim.setup()
    sim.initialize_simulants()
    assert sim.get_number_of_steps_remaining() == 3
    sim.step()
    assert sim.get_number_of_steps_remaining() == 2



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeVar

import pandas as pd

from vivarium import Component
from vivarium.framework.resource import Resource
from vivarium.framework.values.exceptions import DynamicValueError
from vivarium.manager import Manager

if TYPE_CHECKING:
    from vivarium.framework.values.combiners import ValueCombiner
    from vivarium.framework.values.manager import ValuesManager
    from vivarium.framework.values.post_processors import PostProcessor

T = TypeVar("T")


class ValueSource(Resource):
    """A resource representing the source of a value pipeline."""

    def __init__(
        self,
        pipeline: Pipeline,
        source: Callable[..., Any] | None,
        component: Component | None,
    ) -> None:
        super().__init__(
            "value_source" if source else "missing_value_source", pipeline.name, component
        )
        self._pipeline = pipeline
        self._source = source

    def __bool__(self) -> bool:
        return self._source is not None

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if not self._source:
            raise DynamicValueError(
                f"The dynamic value pipeline for {self.name} has no source."
                " This likely means you are attempting to modify a value that"
                " hasn't been created."
            )
        return self._source(*args, **kwargs)


class ValueModifier(Resource):
    """A resource representing a modifier of a value pipeline."""

    def __init__(
        self,
        pipeline: Pipeline,
        modifier: Callable[..., Any],
        component: Component | Manager | None,
    ) -> None:
        mutator_name = self._get_modifier_name(modifier)
        mutator_index = len(pipeline.mutators) + 1
        name = f"{pipeline.name}.{mutator_index}.{mutator_name}"
        super().__init__("value_modifier", name, component)

        self._pipeline = pipeline
        self._source = modifier

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._source(*args, **kwargs)

    @staticmethod
    def _get_modifier_name(modifier: Callable[..., Any]) -> str:
        """Get reproducible modifier names based on the modifier type."""
        if hasattr(modifier, "name"):  # This is Pipeline or lookup table or something similar
            modifier_name: str = modifier.name
        elif hasattr(modifier, "__self__") and hasattr(
            modifier, "__name__"
        ):  # This is a bound method of a component or other object
            owner = modifier.__self__
            owner_name = owner.name if hasattr(owner, "name") else owner.__class__.__name__
            modifier_name = f"{owner_name}.{modifier.__name__}"
        elif hasattr(modifier, "__name__"):  # Some unbound function
            modifier_name = modifier.__name__
        elif hasattr(modifier, "__call__"):  # Some anonymous callable
            modifier_name = f"{modifier.__class__.__name__}.__call__"
        else:  # I don't know what this is.
            raise ValueError(f"Unknown modifier type: {type(modifier)}")
        return modifier_name


class Pipeline(Resource):
    """A tool for building up values across several components.

    Pipelines are lazily initialized so that we don't have to put constraints
    on the order in which components are created and set up. The values manager
    will configure a pipeline (set all of its attributes) when the pipeline
    source is created.

    As long as a pipeline is not actually called in a simulation, it does not
    need a source or to be configured. This might occur when writing
    generic components that create a set of pipeline modifiers for
    values that won't be used in the particular simulation.
    """

    def __init__(self, name: str, component: Component | None = None) -> None:
        super().__init__("value", name, component=component)

        self.source: ValueSource = ValueSource(self, source=None, component=None)
        """The callable source of the value represented by the pipeline."""
        self.mutators: list[ValueModifier] = []
        """A list of callables that directly modify the pipeline source or
        contribute portions of the value."""
        self._combiner: ValueCombiner | None = None
        self.post_processor: PostProcessor | None = None
        """An optional final transformation to perform on the combined output of
        the source and mutators."""
        self._manager: ValuesManager | None = None

    def _get_attr_error(self, attribute: str) -> str:
        return (
            f"The pipeline for {self.name} has no {attribute}. This likely means "
            f"you are attempting to modify a value that hasn't been created."
        )

    def _set_attr_error(self, attribute: str, new_value: Any) -> str:
        current_value = getattr(self, f"_{attribute}")
        return (
            f"A second component is attempting to set the {attribute} for pipeline {self.name} "
            f"with {new_value}, but it already has a {attribute}: {current_value}."
        )

    def _get_property(self, property: T | None, property_name: str) -> T:
        if property is None:
            raise DynamicValueError(self._get_attr_error(property_name))
        return property

    @property
    def combiner(self) -> ValueCombiner:
        """A strategy for combining the source and mutator values into the
        final value represented by the pipeline."""
        return self._get_property(self._combiner, "combiner")

    @property
    def manager(self) -> ValuesManager:
        """A reference to the simulation values manager."""
        return self._get_property(self._manager, "manager")

    def __call__(self, *args: Any, skip_post_processor: bool = False, **kwargs: Any) -> Any:
        """Generates the value represented by this pipeline.

        Arguments
        ---------
        skip_post_processor
            Whether we should invoke the post-processor on the combined
            source and mutator output or return without post-processing.
            This is useful when the post-processor acts as some sort of final
            unit conversion (e.g. the rescale post processor).
        args, kwargs
            Pipeline arguments.  These should be the arguments to the
            callable source of the pipeline.

        Returns
        -------
            The value represented by the pipeline.

        Raises
        ------
        DynamicValueError
            If the pipeline is invoked without a source set.
        """
        return self._call(*args, skip_post_processor=skip_post_processor, **kwargs)

    def _call(self, *args: Any, skip_post_processor: bool = False, **kwargs: Any) -> Any:
        if not self.source:
            raise DynamicValueError(
                f"The dynamic value pipeline for {self.name} has no source. This likely means "
                f"you are attempting to modify a value that hasn't been created."
            )
        value = self.source(*args, **kwargs)
        for mutator in self.mutators:
            value = self.combiner(value, mutator, *args, **kwargs)
        if self.post_processor and not skip_post_processor:
            return self.post_processor(value, self.manager)
        if isinstance(value, pd.Series):
            value.name = self.name

        return value

    def __repr__(self) -> str:
        return f"_Pipeline({self.name})"

    def __hash__(self) -> int:
        return hash(self.name)

    def get_value_modifier(
        self, modifier: Callable[..., Any], component: Component | Manager | None
    ) -> ValueModifier:
        """Add a value modifier to the pipeline and return it.

        Parameters
        ----------
        modifier
            The value modifier callable for the ValueModifier.
        component
            The component that creates the value modifier.
        """
        value_modifier = ValueModifier(self, modifier, component)
        self.mutators.append(value_modifier)
        return value_modifier

    def set_attributes(
        self,
        component: Component | None,
        source: Callable[..., Any],
        combiner: ValueCombiner,
        post_processor: PostProcessor | None,
        manager: ValuesManager,
    ) -> None:
        """
        Add a source, combiner, post-processor, and manager to a pipeline.

        Parameters
        ----------
        component
            The component that creates the pipeline.
        source
            The callable source of the value represented by the pipeline.
        combiner
            A strategy for combining the source and mutator values into the
            final value represented by the pipeline.
        post_processor
            An optional final transformation to perform on the combined output
            of the source and mutators.
        manager
            The simulation values manager.
        """
        self.component = component
        self.source = ValueSource(self, source, component)
        self._combiner = combiner
        self.post_processor = post_processor
        self._manager = manager
